'use strict'

const FORCE_TEST = false || process.env.UPDATE_FORCE_TEST !== undefined
const TEST_FROM_LOCALHOST = false
const useVKUpdater = process.platform == 'win32' || process.platform == 'linux'

let autoUpdater = null
let vkUpdater = null
let bridge = null

if (useVKUpdater) {
  vkUpdater = require('./windows-linux-updater')
  if (FORCE_TEST && false) {
    autoUpdater = new vkUpdater.SelfUpdater()
    autoUpdater.allowHttp = true // TODO updater_worker
  } else {
    autoUpdater = vkUpdater.updater
  }
} else {
  autoUpdater = require('electron').autoUpdater
}

const https = require('https')
const http = require('http')
const URL = require('url')
const settings = require('./settings')
const win = require('./win')
const package_json = require('../package.json')
const tray = require('./tray')
const IPCMainBridge = require('./ipc_main_bridge')
const {getLang} = require('./util')

/**
 * @return {String}
 */
function getPlatform() {
  if (package_json.config.platform) {
    return package_json.config.platform
  }
  if (process.platform == 'darwin') {
    return 'mac'
  }
  if (process.platform == 'win32') {
    return process.arch == 'x64' ? 'win64' : 'win32'
  }
  if (process.platform == 'linux') {
    return process.arch == 'x64' ? 'linux64' : 'linux32'
  }
  return process.platform
}

/**
 * @return {Number}
 */
function getCurrentBuild() {
  if (process.env.UPDATE_CURRENT_BUILD) {
    return parseInt(process.env.UPDATE_CURRENT_BUILD, 10)
  }
  //if (FORCE_TEST) {
  //  return 1
  //}
  return package_json.build
}

/**
 * @return {String}
 */
function getCheckURL() {
  if (TEST_FROM_LOCALHOST) {
    return 'http://localhost/latest.json'
  }

  let url = 'https://desktop.userapi.com/check_update?' +
    'platform=' + getPlatform() +
    '&build=' + getCurrentBuild() +
    '&branch=' + package_json.config.branch +
    '&guid=' + settings.getString('guid') +
    '&lang=' + getLang()

  if (package_json.config.branch != 'master') {
    url += '&secret=' + (package_json.config.secrets.check_update || process.env.CHECK_UPDATE_SECRET || '')
  }

  return url
}

/**
 * @return {Promise} resolves null or update object
 */
function checkForUpdates() {
//  if (FORCE_TEST) {
//    return Promise.resolve(true)
//  }

  let url = getCheckURL()

  return bridge.call({
    name: 'check',
    url
  })
  .then(result => {
    //log('updater: bridge.call(check) result:', result)

    if (result.status != 'update-available') {
      return false
    }

    return result.data
  })
}


module.exports = {
  init() {
    if (!package_json.config.updates) {
      return
    }

    if (!process.env.HOT || FORCE_TEST) {
      autoUpdater.setFeedURL(getCheckURL())
    }

    bridge = new IPCMainBridge('updater_worker')
    bridge.setWindow(win.getWindow())

    bridge.listen((payload, reply) => {
      switch (payload.name) {
      case 'download_progress':
        //log('[bridge] download_progress:', payload.progress)
        autoUpdater.emit('download-progress', payload.progress)
        break
      }
    })

    if (useVKUpdater) {
      autoUpdater.setBridge(bridge)
    }

    autoUpdater.removeAllListeners('checking-for-update')
    autoUpdater.on('checking-for-update', function() {
      //log('main_update.js: checking-for-update')
    })

    autoUpdater.removeAllListeners('update-available')
    autoUpdater.on('update-available', function() {
      log('update: update-available (downloading..)')
    })

    autoUpdater.removeAllListeners('update-not-available')
    autoUpdater.on('update-not-available', function() {
      log('update: update-not-available')
      win.send('update', { status: 'not-available' })
    })

    autoUpdater.removeAllListeners('update-downloaded')
    autoUpdater.on('update-downloaded', function(data) {
      log('update: update-downloaded', data)
      win.send('update', { status: 'downloaded' })
    })

    autoUpdater.removeAllListeners('error')
    autoUpdater.on('error', function(error) {
      log('update: update failed', error)
      win.send('update', { status: 'not-available' })
    })

    if (useVKUpdater) {
      if (process.platform == 'win32') {
        autoUpdater.setExecutableName('vk.exe')
      }

      autoUpdater.removeAllListeners('download-progress')
      autoUpdater.on('download-progress', function(progress) {
        win.send('update', { downloadProgress: progress })
      })
    }
  },

  /**
   * @param {Boolean} byUser
   */
  check(byUser = false) {
    if (!package_json.config.updates) {
      return
    }

    win.send('update', { status: 'checking' })

    checkForUpdates()
    .then(function(update) {
      log('update:', update)
      if (update) {
        win.send('update', { status: 'available', update, byUser })
      } else {
        win.send('update', { status: 'not-available', byUser })
      }
    })
    .catch(function(error) {
      win.send('update', { status: 'not-available', byUser })
    })
  },

  download() {
    log('downloading update')
    autoUpdater.checkForUpdates()
  },

  quitAndInstall() {
    win.getWindow().forceClose = true
    settings.save()
    if (process.platform == 'linux') {
      tray.destroy()
    }
    autoUpdater.quitAndInstall()
  },

  getCheckURL
}
